from typing import Any, Callable, Dict, List, Union

import inspect

from torch.optim.optimizer import Optimizer


def ensure_list(input: Any) -> List:
    """Wrap the input into a list if it is not a list. If it is a None, return an empty list.

    Args:
        input (Any): Input to wrap into a list.

    Returns:
        Output list.
    """
    if isinstance(input, list):
        return input
    if isinstance(input, tuple):
        return list(input)
    if input is None:
        return []
    return [input]


def setattr_dot_notation(obj: Callable, attr: str, value: Any) -> None:
    """Set object's attribute. Supports dot notation.

    Args:
        obj (Callable): Object.
        attr (str): Attribute name of the object.
        value (Any): Value to set the attribute to.
    """
    if "." not in attr:
        if not hasattr(obj, attr):
            raise AttributeError(f"`{get_name(obj, True)}` has no attribute `{attr}`.")
        setattr(obj, attr, value)
    # Solve recursively if the attribute is defined in dot-notation
    else:
        obj_name, attr = attr.split(".", maxsplit=1)
        setattr_dot_notation(getattr(obj, obj_name), attr, value)


def hasarg(_callable: Callable, arg_name: str) -> bool:
    """Check if a function, class, or method has an argument with the specified name.

    Args:
        _callable (Callable): Function, class, or method to inspect.
        arg_name (str): Argument name to check for.

    Returns:
        bool: `True` if the argument if the specified name exists.
    """

    args = inspect.signature(_callable).parameters.keys()
    return arg_name in args


def get_name(_callable: Callable, include_module_name: bool = False) -> str:
    """Get the name of an object, class or function.

    Args:
        _callable (Callable): Object, class or function.
        include_module_name (bool, optional): Whether to include the name of the module from
            which it comes. Defaults to False.

    Returns:
        Name of the given object, class or function.
    """
    name = type(_callable).__name__ if isinstance(_callable, object) else _callable.__name__
    if include_module_name:
        module = type(_callable).__module__ if isinstance(_callable, object) else _callable.__module__
        name = f"{module}.{name}"
    return name


def apply_fns(data: Any, fns: Union[Callable, List[Callable]]) -> Any:
    """Apply a function or a list of functions on the input.

    Args:
        data (Any): Input to apply the function(s) on.
        fns (Union[Callable, List[Callable]]): Function or list of functions to apply on the input.

    Returns:
        Output after applying the function(s).
    """
    for fn in ensure_list(fns):
        data = fn(data)
    return data


def get_optimizer_stats(optimizer: Optimizer) -> Dict[str, float]:
    """
    Extract learning rates and momentum values from an optimizer into a dictionary.

    This function iterates over the parameter groups of the given optimizer and collects
    the learning rate and momentum (or beta values) for each group. The collected values
    are stored in a dictionary with keys formatted to indicate the optimizer type and
    parameter group index (if multiple groups are present).

    Args:
        optimizer (Optimizer): A PyTorch optimizer instance.

    Returns:
        Dict[str, float]: A dictionary containing the learning rates and momentum values
        for each parameter group in the optimizer. The keys are formatted as:
        - "optimizer/{optimizer_class_name}/lr" for learning rates
        - "optimizer/{optimizer_class_name}/momentum" for momentum values
        If there are multiple parameter groups, the keys will include the group index, e.g.,
        "optimizer/{optimizer_class_name}/lr/group1".
    """
    stats_dict = {}
    for group_idx, group in enumerate(optimizer.param_groups):
        lr_key = f"optimizer/{optimizer.__class__.__name__}/lr"
        momentum_key = f"optimizer/{optimizer.__class__.__name__}/momentum"

        # Add group index to the key if there are multiple parameter groups
        if len(optimizer.param_groups) > 1:
            lr_key += f"/group{group_idx+1}"
            momentum_key += f"/group{group_idx+1}"

        # Extracting learning rate
        stats_dict[lr_key] = group["lr"]

        # Extracting momentum or betas[0] if available
        if "momentum" in group:
            stats_dict[momentum_key] = group["momentum"]
        if "betas" in group:
            stats_dict[momentum_key] = group["betas"][0]

    return stats_dict



from typing import Any, Callable, Dict, List, Optional, Union

import warnings

from pydantic import BaseModel, model_validator
from torch.utils.data import Dataset, Sampler
from torchmetrics import MetricCollection

warnings.filterwarnings("ignore", 'Field name "validate" in "ArgsConfigSchema" shadows an attribute in parent "BaseModel"')

BaseModel.model_config["arbitrary_types_allowed"] = True
BaseModel.model_config["hide_input_in_errors"] = True
BaseModel.model_config["extra"] = "forbid"
# ----- Config schema -----


class ArgsConfigSchema(BaseModel):
    fit: Union[Dict[str, Any], str] = {}
    validate: Union[Dict[str, Any], str] = {}
    predict: Union[Dict[str, Any], str] = {}
    test: Union[Dict[str, Any], str] = {}
    lr_find: Union[Dict[str, Any], str] = {}
    scale_batch_size: Union[Dict[str, Any], str] = {}

    @model_validator(mode="after")
    def check_prohibited_args(self):
        prohibited_keys = ["model", "train_loaders", "validation_loaders", "dataloaders", "datamodule"]
        for field in self.model_fields:
            field_value = getattr(self, field)
            if isinstance(field_value, dict):
                found_keys = [key for key in prohibited_keys if key in field_value]
                if found_keys:
                    raise ValueError(
                        f"Found the following prohibited argument(s) in 'args#{field}': "
                        f"{found_keys}. Model and datasets should be defined within the 'system'."
                    )
            elif isinstance(field_value, str) and not (field_value.startswith("%") or field_value.startswith("@")):
                raise ValueError(f"Only dict or interpolators starting with '%' or '@' are allowed for 'args#{field}'.")
        return self


class ConfigSchema(BaseModel):
    requires: Optional[Any] = None
    project: Optional[str] = None
    vars: Dict[str, Any] = {}
    args: ArgsConfigSchema = ArgsConfigSchema()
    system: Dict[str, Any] = {}
    trainer: Dict[str, Any] = {}

    def __init__(self, **data):
        # Annoying workardound, Pydantic keeps all underscored keys private
        requires = data.pop("_requires_") if "_requires_" in data else None
        super().__init__(**data)
        self.requires = requires


# ----- LighterSystem schema -----


class DatasetSchema(BaseModel):
    train: Optional[Dataset] = None
    val: Optional[Dataset] = None
    test: Optional[Dataset] = None
    predict: Optional[Dataset] = None


class SamplerSchema(BaseModel):
    train: Optional[Sampler] = None
    val: Optional[Sampler] = None
    test: Optional[Sampler] = None
    predict: Optional[Sampler] = None


class CollateFnSchema(BaseModel):
    train: Optional[Callable] = None
    val: Optional[Callable] = None
    test: Optional[Callable] = None
    predict: Optional[Callable] = None


class MetricsSchema(BaseModel):
    train: Optional[Union[Any, List[Any], Dict[str, Any]]] = None
    val: Optional[Union[Any, List[Any], Dict[str, Any]]] = None
    test: Optional[Union[Any, List[Any], Dict[str, Any]]] = None

    @model_validator(mode="after")
    def setup_metrics(self):
        for field in self.model_fields:
            if getattr(self, field) is not None:
                setattr(self, field, MetricCollection(getattr(self, field)))
        return self


class ModeSchema(BaseModel):
    train: Optional[Union[Callable, List[Callable]]] = None
    val: Optional[Union[Callable, List[Callable]]] = None
    test: Optional[Union[Callable, List[Callable]]] = None
    predict: Optional[Union[Callable, List[Callable]]] = None


class DataSchema(BaseModel):
    input: Optional[Union[Callable, List[Callable]]] = None
    target: Optional[Union[Callable, List[Callable]]] = None
    pred: Optional[Union[Callable, List[Callable]]] = None


class PostprocessingSchema(BaseModel):
    batch: ModeSchema = ModeSchema()
    criterion: DataSchema = DataSchema()
    metrics: DataSchema = DataSchema()
    logging: DataSchema = DataSchema()



from typing import Any

from functools import partial

import fire
from monai.bundle.config_parser import ConfigParser
from pytorch_lightning import Trainer, seed_everything
from pytorch_lightning.tuner import Tuner

from lighter.system import LighterSystem
from lighter.utils.dynamic_imports import import_module_from_path
from lighter.utils.schema import ArgsConfigSchema, ConfigSchema


def cli() -> None:
    """Defines the command line interface for running lightning trainer's methods."""
    commands = {method: partial(run, method) for method in ArgsConfigSchema.model_fields}
    fire.Fire(commands)


def parse_config(**kwargs) -> ConfigParser:
    """
    Parses configuration files and updates the provided parser
    with given keyword arguments. Returns an updated parser object.

    Args:
        **kwargs (dict): Keyword arguments containing 'config' and, optionally, config overrides.
    Returns:
        An instance of ConfigParser with configuration and overrides merged and parsed.

    Raises:
        ValueError: If '--config' is not specified in the keyword arguments.
        ValueError: If the configuration validation against the schema fails.
    """
    if "config" not in kwargs:
        raise ValueError("'--config' not specified. Please provide a valid configuration file.")

    # Initialize the parser with the predefined structure.
    parser = ConfigParser(ConfigSchema().model_dump(), globals=False)
    # Update the parser with the configuration file.
    parser.update(parser.load_config_files(kwargs.pop("config")))
    # Update the parser with the provided cli arguments.
    parser.update(kwargs)
    # Validate the configuration against the schema.
    ConfigSchema(**parser.config)
    return parser


def run(*methods: str, **kwargs: Any) -> None:
    """
    Execute specified methods of the Lightning Trainer or Tuner.

    Args:
        *methods (str): Method name(s) to be executed.
        **kwargs (Any): Keyword arguments that include 'config' and specific config overrides passed to `parse_config()`.
    """
    for method in methods:
        if method not in ArgsConfigSchema.model_fields:
            valid_methods = list(ArgsConfigSchema.model_fields)
            raise ValueError(
                f"Invalid command '{method}' specified. Available commands: {valid_methods}. "
                f"If you intended to pass an argument, ensure it is prefixed with '--'."
            )

    seed_everything()

    # Parse and validate the config.
    parser = parse_config(**kwargs)

    # Project. If specified, the give path is imported as a module.
    project = parser.get_parsed_content("project")
    if project is not None:
        import_module_from_path("project", project)

    # System
    system = parser.get_parsed_content("system")
    if not isinstance(system, LighterSystem):
        raise ValueError("Expected 'system' to be an instance of 'LighterSystem'")

    # Trainer
    trainer = parser.get_parsed_content("trainer")
    if not isinstance(trainer, Trainer):
        raise ValueError("Expected 'trainer' to be an instance of PyTorch Lightning 'Trainer'")

    # Save the config to checkpoints under "hyper_parameters". Log it if a logger is defined.
    system.save_hyperparameters(parser.config)
    if trainer.logger is not None:
        trainer.logger.log_hyperparams(parser.config)

    # Run the Trainer/Tuner method(s).
    for method in methods:
        args = parser.get_parsed_content(f"args#{method}", default={})
        if hasattr(trainer, method):
            getattr(trainer, method)(system, **args)
        elif hasattr(Tuner, method):
            tuner = Tuner(trainer)
            getattr(tuner, method)(system, **args)



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from functools import partial

import pytorch_lightning as pl
from loguru import logger
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader, Dataset, Sampler
from torchmetrics import Metric, MetricCollection

from lighter.utils.collate import collate_replace_corrupted
from lighter.utils.misc import apply_fns, get_optimizer_stats, hasarg
from lighter.utils.patches import PatchedModuleDict
from lighter.utils.schema import CollateFnSchema, DatasetSchema, MetricsSchema, PostprocessingSchema, SamplerSchema


class LighterSystem(pl.LightningModule):
    """_summary_

    Args:
        model (Module): Model.
        batch_size (int): Batch size.
        drop_last_batch (bool, optional): Whether the last batch in the dataloader should be dropped. Defaults to False.
        num_workers (int, optional): Number of dataloader workers. Defaults to 0.
        pin_memory (bool, optional): Whether to pin the dataloaders memory. Defaults to True.
        optimizer (Optimizer, optional): Optimizers. Defaults to None.
        scheduler (LRScheduler, optional): Learning rate scheduler. Defaults to None.
        criterion (Callable, optional): Criterion/loss function. Defaults to None.
        datasets (Dict[str, Dataset], optional): Datasets for train, val, test, and predict. Defaults to None.
        samplers (Dict[str, Sampler], optional): Samplers for train, val, test, and predict. Defaults to None.
        collate_fns (Dict[str, Union[Callable, List[Callable]]], optional):
            Collate functions for train, val, test, and predict. Defaults to None.
        metrics (Dict[str, Union[Metric, List[Metric], Dict[str, Metric]]], optional):
            Metrics for train, val, and test. Supports a single metric or a list/dict of `torchmetrics` metrics.
            Defaults to None.
        postprocessing (Dict[str, Union[Callable, List[Callable]]], optional):
            Functions to apply to:
                1) The batch returned from the train/val/test/predict Dataset. Defined separately for each.
                2) The input, target, or pred data prior to criterion/metrics/logging. Defined separately for each.

            Follow this structure (all keys are optional):
            ```
                batch:
                    train:
                    val:
                    test:
                    predict:
                criterion / metrics / logging:
                    input:
                    target:
                    pred:
            ```
            Note that the postprocessing of a latter stage stacks on top of the prior ones - for example,
            the logging postprocessing will be done on the data that has been postprocessed for the criterion
            and metrics earlier. Defaults to None.
        inferer (Callable, optional): Inferer must be a class with a `__call__` method that accepts two
            arguments - the input to infer over, and the model itself. Used in 'val', 'test', and 'predict'
            mode, but not in 'train'. Typically, an inferer is a sliding window or a patch-based inferer
            that will infer over the smaller parts of the input, combine them, and return a single output.
            The inferers provided by MONAI cover most of such cases (https://docs.monai.io/en/stable/inferers.html).
            Defaults to None.
    """

    def __init__(
        self,
        model: Module,
        batch_size: int,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        pin_memory: bool = True,
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional[LRScheduler] = None,
        criterion: Optional[Callable] = None,
        datasets: Dict[str, Dataset] = None,
        samplers: Dict[str, Sampler] = None,
        collate_fns: Dict[str, Union[Callable, List[Callable]]] = None,
        metrics: Dict[str, Union[Metric, List[Metric], Dict[str, Metric]]] = None,
        postprocessing: Dict[str, Union[Callable, List[Callable]]] = None,
        inferer: Optional[Callable] = None,
    ) -> None:
        super().__init__()

        # Model setup
        self.model = model
        self.batch_size = batch_size

        # Criterion, optimizer, and scheduler
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler

        # DataLoader specifics
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.drop_last_batch = drop_last_batch

        self.datasets = DatasetSchema(**(datasets or {})).model_dump()
        self.samplers = SamplerSchema(**(samplers or {})).model_dump()
        self.collate_fns = CollateFnSchema(**(collate_fns or {})).model_dump()
        self.postprocessing = PostprocessingSchema(**(postprocessing or {})).model_dump()
        self.metrics = MetricsSchema(**(metrics or {})).model_dump()

        self.metrics = PatchedModuleDict(self.metrics)

        # Inferer for val, test, and predict
        self.inferer = inferer

        # Bypasses LightningModule's check for dataloader and step methods. We define them dynamically in self.setup().
        self.train_dataloader = self.val_dataloader = self.test_dataloader = self.predict_dataloader = lambda: []
        self.training_step = self.validation_step = self.test_step = self.predict_step = lambda: None

    def forward(self, input: Union[Tensor, List[Tensor], Tuple[Tensor], Dict[str, Tensor]]) -> Any:
        """Forward pass. Multi-input models are supported.

        Args:
            input (Tensor, List[Tensor], Tuple[Tensor], Dict[str, Tensor]): Input to the model.

        Returns:
            Output of the model.
        """

        # Keyword arguments to pass to the forward method
        kwargs = {}
        # Add `epoch` argument if forward accepts it
        if hasarg(self.model.forward, "epoch"):
            kwargs["epoch"] = self.current_epoch
        # Add `step` argument if forward accepts it
        if hasarg(self.model.forward, "step"):
            kwargs["step"] = self.global_step

        return self.model(input, **kwargs)

    def configure_optimizers(self) -> Dict:
        """LightningModule method. Returns optimizers and, if defined, schedulers.

        Returns:
            Optimizer and, if defined, scheduler.
        """
        if self.optimizer is None:
            raise ValueError("Please specify 'system.optimizer' in the config.")
        if self.scheduler is None:
            return {"optimizer": self.optimizer}
        else:
            return {"optimizer": self.optimizer, "lr_scheduler": self.scheduler}

    def setup(self, stage: str) -> None:
        """Automatically called by the LightningModule after the initialization.
        `LighterSystem`'s setup checks if the required dataset is provided in the config and
        sets up LightningModule methods for the stage in which the system is.

        Args:
            stage (str): Passed automatically by PyTorch Lightning. ["fit", "validate", "test"].
        """
        # Training methods.
        if stage in ["fit", "tune"]:
            self.train_dataloader = partial(self._base_dataloader, mode="train")
            self.training_step = partial(self._base_step, mode="train")

        # Validation methods. Required in 'validate' stage and optionally in 'fit' or 'tune' stage.
        if stage == "validate" or (stage in ["fit", "tune"] and self.datasets["val"] is not None):
            self.val_dataloader = partial(self._base_dataloader, mode="val")
            self.validation_step = partial(self._base_step, mode="val")

        # Test methods.
        if stage == "test":
            self.test_dataloader = partial(self._base_dataloader, mode="test")
            self.test_step = partial(self._base_step, mode="test")

        # Predict methods.
        if stage == "predict":
            self.predict_dataloader = partial(self._base_dataloader, mode="predict")
            self.predict_step = partial(self._base_step, mode="predict")

    def _base_dataloader(self, mode: str) -> DataLoader:
        """Instantiate the dataloader for a mode (train/val/test/predict).
        Includes a collate function that enables the DataLoader to replace
        None's (alias for corrupted examples) in the batch with valid examples.
        To make use of it, write a try-except in your Dataset that handles
        corrupted data by returning None instead.

        Args:
            mode (str): Mode of operation for which to create the dataloader ["train", "val", "test", "predict"].

        Returns:
            Instantiated DataLoader.
        """
        dataset = self.datasets[mode]
        sampler = self.samplers[mode]
        collate_fn = self.collate_fns[mode]

        if dataset is None:
            raise ValueError(f"Please specify '{mode}' dataset in the 'datasets' key of the config.")

        # Batch size is 1 when using an inference for two reasons:
        # 1) Inferer separates an input into multiple parts, forming a batch of its own.
        # 2) The val/test/pred data usually differ in their shape, so they cannot be stacked into a batch.
        batch_size = self.batch_size
        if self.inferer is not None and mode in ["val", "test", "predict"]:
            logger.info(f"Setting the '{mode}' mode dataloader to batch size of 1 because an inferer is provided.")
            batch_size = 1

        # A dataset can return None when a corrupted example occurs. This collate
        # function replaces None's with valid examples from the dataset.
        collate_fn = partial(collate_replace_corrupted, dataset=dataset, default_collate_fn=collate_fn)
        return DataLoader(
            dataset,
            sampler=sampler,
            shuffle=(mode == "train" and sampler is None),
            batch_size=batch_size,
            drop_last=(self.drop_last_batch if mode == "train" else False),
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            collate_fn=collate_fn,
        )

    def _base_step(self, batch: Dict, batch_idx: int, mode: str) -> Union[Dict[str, Any], Any]:
        """Base step for all modes.

        Args:
            batch (Dict): Batch data as a containing "input", and optionally "target" and "id".
            batch_idx (int): Batch index. PyTorch Lightning requires it, even though it is not used here.
            mode (str): Operating mode. (train/val/test/predict)

        Returns:
            For the predict step, it returns pred only.
            For the training, validation, and test steps, it returns a dictionary
            containing loss, metrics, input, target, pred, and id. Loss is `None`
            for the test step. Metrics is `None` if no metrics are specified.
        """
        # Allow postprocessing on batch data. Can be used to restructure the batch data into the required format.
        batch = apply_fns(batch, self.postprocessing["batch"][mode])

        # Verify that the batch is a dict with valid keys.
        batch_err_msg = "Batch must be a dict with keys:\n\t- 'input'\n\t- 'target' (optional)\n\t- 'id' (optional)\n"
        if not isinstance(batch, dict):
            raise TypeError(batch_err_msg + f"Batch type found: '{type(batch).__name__}'.")
        if set(batch.keys()) not in [{"input"}, {"input", "target"}, {"input", "id"}, {"input", "target", "id"}]:
            raise ValueError(batch_err_msg + f"Batch keys found: {batch.keys()}.")

        # Ensure that the returned values of the optional batch keys are not None.
        for key in ["target", "id"]:
            if key in batch and batch[key] is None:
                raise ValueError(f"Batch's '{key}' value cannot be None. If '{key}' should not exist, omit it.")

        # Unpack the batch. The target and id are optional. If not provided, they are set to None for internal use.
        input = batch["input"]
        target = batch.get("target", None)
        id = batch.get("id", None)

        # Forward
        if self.inferer and mode in ["val", "test", "predict"]:
            pred = self.inferer(input, self)
        else:
            pred = self(input)

        # Postprocessing for loss calculation.
        input = apply_fns(input, self.postprocessing["criterion"]["input"])
        target = apply_fns(target, self.postprocessing["criterion"]["target"])
        pred = apply_fns(pred, self.postprocessing["criterion"]["pred"])

        # Predict mode stops here.
        if mode == "predict":
            # Postprocessing for logging/writing.
            pred = apply_fns(pred, self.postprocessing["logging"]["pred"])
            return {"pred": pred, "id": id}

        # Calculate the loss.
        loss = None
        if mode in ["train", "val"]:
            # When target is not provided, pass only the predictions to the criterion.
            loss = self.criterion(pred) if target is None else self.criterion(pred, target)

        # Postprocessing for metrics.
        input = apply_fns(input, self.postprocessing["metrics"]["input"])
        target = apply_fns(target, self.postprocessing["metrics"]["target"])
        pred = apply_fns(pred, self.postprocessing["metrics"]["pred"])

        # Calculate the step metrics.
        metrics = None
        if self.metrics[mode] is not None:
            metrics = self.metrics[mode](pred, target)

        # Postprocessing for logging/writing.
        input = apply_fns(input, self.postprocessing["logging"]["input"])
        target = apply_fns(target, self.postprocessing["logging"]["target"])
        pred = apply_fns(pred, self.postprocessing["logging"]["pred"])

        # Ensure that a dict of losses has a 'total' key.
        if isinstance(loss, dict) and "total" not in loss:
            raise ValueError(
                "The loss dictionary must include a 'total' key that combines all sublosses. "
                "Example: {'total': combined_loss, 'subloss1': loss1, ...}"
            )

        # Logging
        self._log_stats(loss, metrics, mode, batch_idx)

        # Return the loss as required by Lightning as well as other data that can be used in hooks or callbacks.
        return {
            "loss": loss["total"] if isinstance(loss, dict) else loss,
            "metrics": metrics,
            "input": input,
            "target": target,
            "pred": pred,
            "id": id,
        }

    def _log_stats(self, loss: Union[Tensor, Dict[str, Tensor]], metrics: MetricCollection, mode: str, batch_idx: int) -> None:
        """
        Logs the loss, metrics, and optimizer statistics.
        Args:
            loss (Union[Tensor, Dict[str, Tensor]]): Calculated loss or a dict of sublosses.
            metrics (MetricCollection): Calculated metrics.
            mode (str): Mode of operation (train/val/test/predict).
            batch_idx (int): Index of current batch.
        """
        if self.trainer.logger is None:
            return

        on_step_log = partial(self.log, logger=True, batch_size=self.batch_size, on_step=True, on_epoch=False, sync_dist=False)
        on_epoch_log = partial(self.log, logger=True, batch_size=self.batch_size, on_step=False, on_epoch=True, sync_dist=True)

        # Loss
        if loss is not None:
            if not isinstance(loss, dict):
                on_step_log(f"{mode}/loss/step", loss)
                on_epoch_log(f"{mode}/loss/epoch", loss)
            else:
                for name, subloss in loss.items():
                    on_step_log(f"{mode}/loss/{name}/step", subloss)
                    on_epoch_log(f"{mode}/loss/{name}/epoch", subloss)
        # Metrics
        if metrics is not None:
            for name, metric in metrics.items():
                on_step_log(f"{mode}/metrics/{name}/step", metric)
                on_epoch_log(f"{mode}/metrics/{name}/epoch", metric)
        # Optimizer's lr, momentum, beta. Logged in train mode and once per epoch.
        if mode == "train" and batch_idx == 0:
            for name, optimizer_stat in get_optimizer_stats(self.optimizer).items():
                on_epoch_log(f"{mode}/{name}", optimizer_stat)

    @property
    def learning_rate(self) -> float:
        """Get the learning rate of the optimizer. Ensures compatibility with the Tuner's 'lr_find()' method."""
        if len(self.optimizer.param_groups) > 1:
            raise ValueError("The learning rate is not available when there are multiple optimizer parameter groups.")
        return self.optimizer.param_groups[0]["lr"]

    @learning_rate.setter
    def learning_rate(self, value) -> None:
        """Set the learning rate of the optimizer. Ensures compatibility with the Tuner's 'lr_find()' method."""
        if len(self.optimizer.param_groups) > 1:
            raise ValueError("The learning rate is not available when there are multiple optimizer parameter groups.")
        self.optimizer.param_groups[0]["lr"] = value
